/*
    saa7110 - Philips SAA7110(A) video decoder driver

    Copyright (C) 1998 Pauline Middelink <middelin@polyware.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include <linux/i2c.h>
#include <linux/videodev.h>
#include "video_decoder.h"

#define DEBUG(x...)			/* remove when no long debugging */

#define SAA7110_MAX_INPUT	9	/* 6 CVBS, 3 SVHS */
#define SAA7110_MAX_OUTPUT	0	/* its a decoder only */

#define	I2C_SAA7110		0x9C	/* or 0x9E */

#define	I2C_DELAY		10	/* 10 us or 100khz */

struct saa7110 {
	int	channel;
	int	norm;
	int     bright;
	int     contrast;
	int     hue;
	int     sat;
};

/* ----------------------------------------------------------------------- */
/* I2C support functions						   */
/* ----------------------------------------------------------------------- */
static
int saa7110_write(struct i2c_device *dev, unsigned char subaddr, unsigned char data)
{
	int ack;
	unsigned long flags;

	LOCK_I2C_BUS(dev->bus);
	i2c_start(dev->bus);
	i2c_sendbyte(dev->bus, dev->addr, I2C_DELAY);
	i2c_sendbyte(dev->bus, subaddr, I2C_DELAY);
	ack = i2c_sendbyte(dev->bus, data, I2C_DELAY);
	i2c_stop(dev->bus);
	UNLOCK_I2C_BUS(dev->bus);
	return ack;
}

/* ----------------------------------------------------------------------- */
/* SAA7110 functions							   */
/* ----------------------------------------------------------------------- */
static
int saa7110_selmux(struct i2c_device *device, int chan)
{
static	const unsigned char modes[9][8] = {
/* mode 1 */	{ 0x00, 0xD9, 0x17, 0x40, 0x03, 0x44, 0x75, 0x16 },
/* mode 2 */	{ 0x00, 0xD8, 0x17, 0x40, 0x03, 0x44, 0x75, 0x16 },
/* mode 3 */	{ 0x00, 0xBA, 0x07, 0x91, 0x03, 0x60, 0xB5, 0x05 },
/* mode 4 */	{ 0x00, 0xB8, 0x07, 0x91, 0x03, 0x60, 0xB5, 0x05 },
/* mode 5 */	{ 0x00, 0x7C, 0x07, 0xD2, 0x83, 0x60, 0xB5, 0x03 },
/* mode 6 */	{ 0x00, 0x78, 0x07, 0xD2, 0x83, 0x60, 0xB5, 0x03 },
/* mode 7 */	{ 0x80, 0x59, 0x17, 0x42, 0xA3, 0x44, 0x75, 0x12 },
/* mode 8 */	{ 0x80, 0x9A, 0x17, 0xB1, 0x13, 0x60, 0xB5, 0x14 },
/* mode 9 */	{ 0x80, 0x3C, 0x27, 0xC1, 0x23, 0x44, 0x75, 0x21 } };
	struct i2c_bus* bus = device->bus;
	const unsigned char* ptr;
	int status;

	ptr = modes[chan];
	saa7110_write(device,0x06,ptr[0]);
	saa7110_write(device,0x20,ptr[1]);
	saa7110_write(device,0x21,ptr[2]);
	saa7110_write(device,0x22,ptr[3]);
	saa7110_write(device,0x2C,ptr[4]);
	saa7110_write(device,0x30,ptr[5]);
	saa7110_write(device,0x31,ptr[6]);
	saa7110_write(device,0x21,ptr[7]);

	/* mode changed, start automatic detection */
	mdelay(150);
	status = i2c_read(bus,device->addr|1);
	if ((status & 3) == 0) {
		if (status & 0x20) {
			DEBUG(printk(KERN_DEBUG "zoran: standard=bw60\n"));
			saa7110_write(device,0x06,0x80);
			saa7110_write(device,0x2E,0x81);
			return 0;
		}
		DEBUG(printk(KERN_DEBUG "zoran: standard=bw50\n"));
		saa7110_write(device,0x06,0x80);
		saa7110_write(device,0x2E,0x9A);
		return 0;
	}

	if (status & 0x20) {
		DEBUG(printk(KERN_DEBUG "zoran: standard=ntsc\n"));
		saa7110_write(device,0x0D,0x06);
		saa7110_write(device,0x11,0x2C);
		saa7110_write(device,0x2E,0x81);
		return 0;
	}

	/* PAL/SECAM */
	saa7110_write(device,0x0D,0x06);
	saa7110_write(device,0x11,0x59);
	saa7110_write(device,0x2E,0x9A);

	mdelay(150);	/* pause 150 ms */

	if (0 && (status & 0x03) == 0x01) {
		DEBUG(printk(KERN_DEBUG "zoran: standard=secam\n"));
		saa7110_write(device,0x0D,0x07);
		return 0;
	}
	DEBUG(printk(KERN_DEBUG "zoran: standard=pal\n"));
	return 0;
}

static
int saa7110_attach(struct i2c_device *device)
{
static	const unsigned char initseq[] = {
	     0, 0x4C, 0x3C, 0x0D, 0xEF, 0xBD, 0xF0, 0x00, 0x00,
		0xF8, 0xF8, 0x60, 0x60, 0x00, 0x06, 0x18, 0x90,
		0x00, 0x2C, 0x40, 0x46, 0x42, 0x1A, 0xFF, 0xDA,
		0xF0, 0x8B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xD9, 0x17, 0x40, 0x41, 0x80, 0x41, 0x80, 0x4F,
		0xFE, 0x01, 0xCF, 0x0F, 0x03, 0x01, 0x81, 0x03,
		0x44, 0x75, 0x01, 0x8C, 0x03};
	const unsigned char * const last = initseq+sizeof(initseq);
	const unsigned char* ptr;
	struct i2c_bus* bus = device->bus;

	DEBUG(printk(KERN_DEBUG "saa7110_attach\n"));
	strcpy(device->name, "saa7110");

	device->data = kmalloc(sizeof(struct saa7110), GFP_KERNEL);
	if (device->data == 0)
		return -ENOMEM;

	MOD_INC_USE_COUNT;

	/* setup the chip */
	ptr = initseq;
	i2c_start(bus);
	i2c_sendbyte(bus,device->addr,0);
	while (ptr != last) {
		if (i2c_sendbyte(bus,*ptr++,0)) {
			i2c_stop(bus);
			DEBUG(printk(KERN_DEBUG "zoran: %s init failed on bus %s, driver %s\n", device->name, bus->name, device->driver->name));
			return -EAGAIN;
		}
	}
	i2c_stop(bus);
	saa7110_write(device,0x21,0x16);

	/* clear our private data */
	memset(device->data, 0, sizeof(struct saa7110));

	/* setup and implicit mode 0 select has been performed */
	return 0;
}

static
int saa7110_detach(struct i2c_device *device)
{
	DEBUG(printk(KERN_DEBUG "saa7110_detach\n"));

	/* stop further output */
	saa7110_write(device,0x0E,0x00);

	kfree(device->data);

	MOD_DEC_USE_COUNT;
	return 0;
}

static
int saa7110_command(struct i2c_device *device, unsigned int cmd, void *arg)
{
	struct saa7110* pdev = device->data;
	int	v;

	switch (cmd) {
	 case DECODER_GET_CAPABILITIES:
		{
			struct video_decoder_capability dc;
			dc.flags = VIDEO_DECODER_PAL
				 | VIDEO_DECODER_NTSC
				 | VIDEO_DECODER_SECAM
				 | VIDEO_DECODER_AUTO
				 | VIDEO_DECODER_CCIR;
			dc.inputs = SAA7110_MAX_INPUT;
			memcpy(arg,&dc,sizeof(dc));
		}
		break;

	 case DECODER_GET_NORM:
		*(int*)arg = pdev->norm;
		break;
	 case DECODER_SET_NORM:
		pdev->norm = *(int*)arg;
		DEBUG(printk(KERN_DEBUG "saa7110: set norm(%d)\n",pdev->norm));
		return -EPERM;

	 case DECODER_GET_INPUT:
		*(int*)arg = pdev->channel;
		break;
	 case DECODER_SET_INPUT:
		v = *(int*)arg;
		if (v<0 || v>SAA7110_MAX_INPUT)
			return -EINVAL;
		DEBUG(printk(KERN_DEBUG "saa7110: set input(%d)\n",v));
		return saa7110_selmux(device, v);

	 case DECODER_GET_PICTURE:
		{
			struct video_picture *pic = arg;

			pic->brightness = pdev->bright;
			pic->hue = pdev->hue;
			pic->colour = pdev->sat;
			pic->contrast = pdev->contrast;
			pic->whiteness = 0;
			pic->depth = 0;
			pic->palette = VIDEO_PALETTE_YUYV;
		}
		break;
	 case DECODER_SET_PICTURE:
		{
			struct video_picture *pic = arg;

			if (pdev->bright != pic->brightness) {
				/* We want 0 to 255 we get 0-65535 */
				pdev->bright = pic->brightness;
				saa7110_write(device, 0x19, pdev->bright >> 8);
			}
			if (pdev->contrast != pic->contrast) {
				/* We want 0 to 127 we get 0-65535 */
				pdev->contrast = pic->contrast;
				saa7110_write(device, 0x13, pdev->contrast >> 9);
			}
			if (pdev->sat != pic->colour) {
			/* We want 0 to 127 we get 0-65535 */
				pdev->sat = pic->colour;
				saa7110_write(device, 0x12, pdev->sat >> 9);
			}
			if (pdev->hue != pic->hue) {
				/* We want -128 to 127 we get 0-65535 */
				pdev->hue = pic->hue;
				saa7110_write(device, 0x07, (pdev->hue>>8)-128);
			}
		}
		break;

	 default:
		DEBUG(printk(KERN_DEBUG "unknown saa7110_command??(%d)\n",cmd));
		return -EINVAL;
	}
	return 0;
}

/* ----------------------------------------------------------------------- */

struct i2c_driver i2c_driver_saa7110 =
{
	"saa7110",			/* name */

	I2C_DRIVERID_DECODER,		/* in i2c.h */
	I2C_SAA7110, I2C_SAA7110+1,	/* Addr range */

	saa7110_attach,
	saa7110_detach,
	saa7110_command
};

EXPORT_NO_SYMBOLS;

#ifdef MODULE
int init_module(void)
#else
int saa7110_init(void)
#endif
{
	return i2c_register_driver(&i2c_driver_saa7110);
}

#ifdef MODULE
void cleanup_module(void)
{
	i2c_unregister_driver(&i2c_driver_saa7110);
}
#endif
